#!/bin/bash

############################################################################
#
# Copyright (c) 2006 Mellanox Technologies. All rights reserved.
#
# This Software is licensed under one of the following licenses:
#
# 1) under the terms of the "Common Public License 1.0" a copy of which is
#    available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/cpl.php.
#
# 2) under the terms of the "The BSD License" a copy of which is
#    available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/bsd-license.php.
#
# 3) under the terms of the "GNU General Public License (GPL) Version 2" a
#    copy of which is available from the Open Source Initiative, see
#    http://www.opensource.org/licenses/gpl-license.php.
#
# Licensee has the right to choose one of the above licenses.
#
# Redistributions of source code must retain the above copyright
# notice and one of the license notices.
#
# Redistributions in binary form must reproduce both the above copyright
# notice, one of the license notices in the documentation
# and/or other materials provided with the distribution.
#
############################################################################

############################################################################
# Set the active slave according to the vnic priority - 
# Lower value - higher priority
############################################################################

# param - vnic I/F
# return values: 0 - not a vnic
#		1 - is a vnic
is_vnic()
{
	if [ -z `$vnic_info -l | grep $1` ]; then
		[ "$VERBOSE" = no ] || echo $1 " is not an EoIB I/F" >> $debug_file
		return 0;
	fi

	[ "$VERBOSE" = no ] || echo $1 " is an EoIB I/F" >> $debug_file
	return 1;
}

# param - bond to get active slave for
get_active_slave()
{
	echo `cat /sys/class/net/$1/bonding/active_slave`
}

# param - vnic I/F
get_vnic_prio()
{
#	echo "info for vnic $1 is: " `$vnic_info -p $1`
#	echo "so prio is: "  `$vnic_info -p $1 |  /bin/awk '{print $2}'`
	a=`$vnic_info -p $1 |  /bin/awk '{print $2}'`
#	echo a: $a
	echo $a
}

# if a member of bond - echo the bond
# return  values: 0 if not a member of bond
#		  1 if a member of bond
get_bond_master()
{
	vbond=`grep MASTER $ifcfg_file | cut -d"=" -f2`
	if [ -z `grep "SLAVE=yes" $ifcfg_file` ] || [ -z $vbond ]; then
		return 0;
	fi

	echo $vbond
	return 1;
}

# param - bond
set_max_prio_vnic_in_bond()
{
	max_prio=-1
	vbond=$1

	#Find the other slaves (vNics) that belong to the same bond as this vNic
	#todo - what if there is no active slave? what if there are no slaves?
	for slave in `cat /sys/class/net/$vbond/bonding/slaves`; do
		# Make sure it is a vnic
		is_vnic $slave
		is_vnic_if=$?
		if [ $is_vnic_if -eq 0 ]; then
			continue;
		fi
		[ "$VERBOSE" = no ] || echo "vnic " $slave " is also a member of bond" $vbond >> $debug_file
		
		vnic_prio=$(get_vnic_prio $slave)
		echo "prio of " $slave " is: " $vnic_prio
		if [ $max_prio -eq -1 ] || [ $vnic_prio -lt $max_prio ]; then
			max_prio=$vnic_prio
			max_prio_vnic=$slave
		fi

	done

	# Check if the active slave should be updated.
	active_slave=$(get_active_slave $vbond)
	echo "vbond : " $vbond
	curr_prio=$(get_vnic_prio $active_slave)
	is_vnic $active_slave
	is_active_vnic=$?
	if [ $max_prio -lt $curr_prio ] || [ $is_active_vnic -eq 0 ]; then
		echo $max_prio_vnic > /sys/class/net/$vbond/bonding/active_slave
		[ "$VERBOSE" = no ] || echo "vnic " $max_prio_vnic " is the active_slave of bond " $vbond >> $debug_file
	fi
}

VERBOSE=${VERBOSE:-"no"}

action=$1
vif=$2

vnics_file=/tmp/vnics_file
debug_file=/tmp/mlx4_vnic_bond_debug
vnic_info=/sbin/mlx4_vnic_info
ifcfg_prefix=/etc/sysconfig/network-scripts/ifcfg-
ifcfg_file=$ifcfg_prefix$vif

# Verify that the mlx4_vnic_info script is found at the expected location
if [ ! -x  $vnic_info ]; then
	echo $vnic_info " script not found"
	exit 1
else 
	echo  $vnic_info " script OK"
fi

is_valid_vnic()
{
	if [ $action = "add"]; then
		is_vnic $vif
		vif_vnic=$?
	else
		vif_vnic=0
		grep -x $vif $vnics_file
		[ $? -eq 0 ] || vif_vnic=1
	fi
	echo $if_vnic	
}
		
# We only work on mlnx EoIB vnics
is_vif_vnic=$(is_valid_vnic)
if [ $is_vif_vnic -eq 0 ]; then
	[ "$VERBOSE" = no ] || echo $vif " is not an EoIB I/F" >> $debug_file
	exit 0
fi

# Look for the relevant ifcfg file, to see to which bond device
# this vnic belongs
if [ ! -f  $ifcfg_file ]; then
	[ "$VERBOSE" = no ] || echo "ifcfg file " $ifcfg_file " for I/F " $vif " not found" >> $debug_file
	exit 0;
fi

vbond=$(get_bond_master)
if [ $? -eq 0 ]; then
	[ "$VERBOSE" = no ] || echo $vif " is not a member of a bond" >> $debug_file
	exit 0;
fi
[ "$VERBOSE" = no ] || echo $vif " is a member of a bond" $vbond >> $debug_file

active_slave=$(get_active_slave $vbond)

case $action in
	"add" )
		# Write the new vnic to a file
		echo $vif >> $vnics_file
		# if this vnic is added - check if it's prio is higher then the bond master and if it is - set as master
		curr_prio=$(get_vnic_prio $active_slave)
		new_prio=$(get_vnic_prio $vif)

		if [ $new_prio -lt $curr_prio ]; then
			[ "$VERBOSE" = no ] || echo "$vif prio $new_prio < $curr_prio curr_prio" >> $debug_file
			echo $vif > /sys/class/net/$vbond/bonding/active_slave
		fi

	;;
	"remove" )
		# if this vnic is being removed - check if it was the bond master and if it was - set another one
		# todo: add a check. the current check is invalid since the vnic was already removed and thus is
		# 	not the active slave
#		if  [ "$active_slave " = "$vif" ]; then
#			[ "$VERBOSE" = no ] || echo "$vif is the active slave in bond $vbond and is being removed" >> $debug_file
			set_max_prio_vnic_in_bond $vbond
#		fi
		sed "/$vif$/d" $vnics_file > /tmp/tmp
		mv /tmp/tmp $vnics_file
		;;
	* )
		[ "$VERBOSE" = no ] || echo "vnic: $vif, bond: $vbond: action $action is not supported by mlx4_vnic_bond" >> $debug_file
		exit 0;
        	;;
esac                                      

[ "$VERBOSE" = no ] || echo "$(get_active_slave $vbond) is the active slave in bond $vbond" >> $debug_file
exit 0
